﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Drawing.Drawing2D;
using System.Threading;
using System.IO;

using VidyaVitamin.Plugin;
using ComRes = VidyaVitamin.Plugin.Properties.Resources;

namespace VidyaVitamin.Plugin.Thumbnail
{
	public partial class ThumbForm : Form, IVVForm, INotifyUpdate, IReceiveCard
	{
		private IFirebirdTable table;
		private IVVUtility util;
		private bool embed;
		private string pct_name;
		private ThumbnailData[] imagecache;
		private List<int> selectedindex = new List<int>();
		private Dictionary<int, ThumbnailData> datacache;

		public ToolStrip MainToolStrip
		{
			get { return tool; }
		}

		public ToolStrip SubToolStrip
		{
			get { return null; }
		}

		public StatusStrip StatusStrip
		{
			get { return null; }
		}

		public TreeNode ProjectNode
		{
			get { return Tag as TreeNode; }
		}

		public EnableType EnableMenus
		{
			get
			{
				EnableType type = EnableType.None;
				if (selectedindex.Count > 0) type |= EnableType.Print;
				if (selectedindex.Count > 0) type |= EnableType.Save;
				return type;
			}
		}

		public void OnClickPrint()
		{
			PrintForm pf = new PrintForm(table, pct_name, imagecache, selectedindex, util);
			pf.ShowDialog();
		}

		public void OnClickSave(bool saveas)
		{
			int count = selectedindex.Count;

			string sqlcmd;
			if (count == 1) {
				int index = imagecache[selectedindex[0]].index;
				sqlcmd = string.Format(
					"SELECT {0} FROM {1} WHERE {2}={3};",
					pct_name, table.TableName, Thumbnail.STR_ID, index);
				string path = table.IFirebird.ExecScalar(sqlcmd) as string;
				if (Path.IsPathRooted(path) == false) path =
					 util.ToAbsolutePath(table.EditPath, path);
				SaveFileDialog dlg = new SaveFileDialog();
				dlg.FileName = path;
				dlg.Filter = ComRes.str_AllFilesFilter;
				dlg.InitialDirectory = Path.GetDirectoryName(path);
				if (dlg.ShowDialog() == DialogResult.Cancel) return;

				if (embed == false) {
					File.Copy(path, dlg.FileName);
				} else {
					sqlcmd = string.Format("SELECT {0} FROM {1}${2} WHERE {3}={4};",
						Thumbnail.STR_BINARY, table.TableName, pct_name, Thumbnail.STR_ID, index);
					File.WriteAllBytes(dlg.FileName, table.IFirebird.ExecScalar(sqlcmd) as byte[]);
				}
			} else if (count > 1) {
				FolderBrowserDialog dlg = new FolderBrowserDialog();
				dlg.Description = Properties.Resources.strSaveFolder;
				dlg.SelectedPath = Path.GetDirectoryName(table.EditPath);
				if(dlg.ShowDialog() == DialogResult.Cancel) return;
				string folder = dlg.SelectedPath;
				if (folder.EndsWith(Path.PathSeparator.ToString()) == false)
					folder += Path.PathSeparator.ToString();
				ProgressForm pf = new ProgressForm(table, pct_name, folder, imagecache, selectedindex);
				pf.ShowDialog();
			}
		}

		private ToolStripTrackBar tsTrackSize;

		public ThumbForm(TreeNode pnode, IFirebirdTable table, IVVUtility util)
		{
			this.util = util;
			this.Tag = pnode;
			this.table = table;

			InitializeComponent();

			tsTrackSize = new ToolStripTrackBar();
			tsTrackSize.Value = 128;
			tsTrackSize.Size = new Size(120, 22);
			tsTrackSize.ValueChanged += new EventHandler(tsSize_ValueChanged);
			tool.Items.Add(tsTrackSize);

			tsTxtKeyword.Size = new Size(80, 22);

			tsBtnQuick.Image = ComRes.img_Find;

			imagecache = null;
			datacache = null;

			Dictionary<DBAttrType, string> pa = table.GetAttributedParts
				(DBAttrType.Picture | DBAttrType.Main | DBAttrType.Category | DBAttrType.Title);
			if (pa.ContainsKey(DBAttrType.Picture) == false) {
				// 画像属性無し
				tool.Enabled = false;
				vsb.Enabled = false;
			} else {
				pct_name = pa[DBAttrType.Picture];
				embed = table.IFirebird.ExistTable(table.TableName + "$" + pct_name);
			}

			// メニューの初期化
			ToolStripMenuItem mi;
			foreach(IVVCardFormPlugin ic in util.EnumPlugins<IVVCardFormPlugin>()){
				if (ic is IVVReceivable) {
					mi = new ToolStripMenuItem(string.Format(ComRes.str_Transfer, ic.PluginName));
					mi.Tag = ic;
					mi.DropDownOpening += new EventHandler(OnOpeningReciveForm);

					ToolStripItem si = mi.DropDownItems.Add(ComRes.str_New);
					si.Click += new EventHandler(OnClickSendNewForm);

					cmenu.Items.Add(mi);
				}
			}

			// クイック検索が使えるか
			if (pa.ContainsKey(DBAttrType.Main) == false
			&& pa.ContainsKey(DBAttrType.Category) == false
			&& pa.ContainsKey(DBAttrType.Title) == false) {
				tsTxtKeyword.Enabled = false;
				tsBtnQuick.Enabled = false;
			}
			tsBtnQuick.Image = ComRes.img_Find;

			UpdateThumbnail();
		}

		private void OnOpeningReciveForm(object sender, EventArgs e)
		{
			ToolStripMenuItem mi = (ToolStripMenuItem)sender;
			ToolStripItemCollection sc = mi.DropDownItems;
			while (sc.Count > 1) sc.RemoveAt(1);

			IVVReceivable ir = mi.Tag as IVVReceivable;
			Type ftype = ir.FormType;

			foreach (IReceiveCard rc in util.EnumTarget<IReceiveCard>((TreeNode)this.Tag, true)) {
				if (rc == this) continue;
				if (rc.GetType().Equals(ftype) == true) {
					ToolStripItem i = sc.Add(((Control)rc).Text);
					i.Click += new EventHandler(OnClickSendExistForm);
					i.Tag = rc;
				}
			}

			if (sc.Count > 1) sc.Insert(1, new ToolStripSeparator());
		}

		private int[] GetSelectedIndexes()
		{
			int count = selectedindex.Count;
			int[] indexes = new int[count];
			int j = count - 1;
			foreach (int i in selectedindex) {
				indexes[j--] = imagecache[i].index;
			}
			return indexes;
		}

		private void OnClickSendNewForm(object sender, EventArgs e)
		{
			TreeNode tnode = ((TreeNode)Tag).Parent;
			ToolStripMenuItem mi = (ToolStripMenuItem)sender;

			IVVCardFormPlugin iv = mi.OwnerItem.Tag as IVVCardFormPlugin;
			if (iv == null) return;
			IVVReceivable ir = iv as IVVReceivable;

			int[] indexes = GetSelectedIndexes();
			if (ir.SingleReceive == false) {
				IVVForm f = util.AppendNewPluginForm(table, tnode, iv);
				if (f is IReceiveCard) ((IReceiveCard)f).ReceiveCard(indexes);
			} else {
				int count = indexes.Length; if (count > 10) count = 10;
				for (int n = 0; n < count; n++) {
					IVVForm f = util.AppendNewPluginForm(table, tnode, iv);
					if (f is IReceiveCard) ((IReceiveCard)f).ReceiveCard(new int[] { indexes[n] });
				}
			}
		}

		private void OnClickSendExistForm(object sender, EventArgs e)
		{
			ToolStripItem si = (ToolStripItem)sender;
			IReceiveCard rc = si.Tag as IReceiveCard;
			if (rc == null) return;

			rc.ReceiveCard(GetSelectedIndexes());
		}

		public void ReceiveCard(int[] indexes)
		{
			// 既存のスレッドをすべて終了
			List<int> dellist = new List<int>();
			foreach (KeyValuePair<int, Thread> p in threaddata) {
				dellist.Add(p.Key);
			}
			foreach (int i in dellist) {
				if (threaddata.ContainsKey(i) == true) threaddata[i].Abort();
			}

			imagecache = new ThumbnailData[indexes.Length];
			datacache = new Dictionary<int,ThumbnailData>();
			for (int i = 0; i < indexes.Length; i++) {
				imagecache[i] = new ThumbnailData();
				imagecache[i].loadfailed = false;
				imagecache[i].index = indexes[i];
				imagecache[i].selected = false;
				datacache.Add(indexes[i], imagecache[i]);
			}
			UpdateThumbnail();
		}

		public void NotifyUpdate(UpdateType type, IFirebirdTable ftable, params object[] param)
		{
			if (type == UpdateType.UpdateCard || type == UpdateType.DeleteCard) {
				if (ftable != table) return;

				ThumbnailData data; int id;
				id = (int)param[0];
				data = datacache[id];
				if (type == UpdateType.UpdateCard) {
					data.image = null;
					data.loadfailed = false;
				} else {
					data.image = null;
					data.loadfailed = true;
				}
				Invalidate();
			}
		}

		private int GetIndexFromPosition(Point pt)
		{
			int w, h;
			GetThumbSize(out w, out h);

			int cw = (ClientSize.Width - padding * 2);
			int wc = cw / w;
			int vp = vsb.Value + pt.Y - padding;
			if (vp < 0) return -1;
			int sy = vp / h * wc;

			int sx = (pt.X - padding);
			if (sx < 0) return -1;
			sx = sx / w;
			if (sx >= wc) return -1;
			return sy + sx;
		}

		protected override void OnMouseDown(MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Right) {
				cmenu.Show(PointToScreen(e.Location));
			}

			base.OnMouseDown(e);
		}

		protected override void OnMouseClick(MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left) {
				int pos = GetIndexFromPosition(e.Location);
				if (pos != -1) {
					if ((ModifierKeys & Keys.Shift) != 0) {
						int c = selectedindex.Count;
						for (int i = 0; i < c - 1; i++) selectedindex.RemoveAt(0);

						int lpos = selectedindex[0];
						if (pos > lpos) {
							int n = pos; pos = lpos; lpos = n;
						}
						selectedindex.Clear();
						for (int i = pos; i <= lpos; i++) {
							imagecache[i].selected = true;
							selectedindex.Add(i);
						}
					} else if ((ModifierKeys & Keys.Control) == 0) {
						foreach (int i in selectedindex) {
							imagecache[i].selected = false;
						}
						selectedindex.Clear();

						imagecache[pos].selected = true;
						selectedindex.Add(pos);
					} else {
						imagecache[pos].selected = !imagecache[pos].selected;
						selectedindex.Add(pos);
					}

					Invalidate();
				}
			}

			base.OnMouseClick(e);
		}

		protected override void OnMouseWheel(MouseEventArgs e)
		{
			base.OnMouseWheel(e);

			int v = vsb.Value;
			v += -e.Delta;
			if (v < vsb.Minimum) v = vsb.Minimum;
			else if (v > vsb.Maximum) v = vsb.Maximum;
			vsb.Value = v;
		}

		private const int margin = 3, textheight = 16, padding = 5;

		private void GetThumbSize(out int w, out int h)
		{
			w = tsTrackSize.Value + margin * 2;
			h = w + textheight;
		}

		private bool updatelock = false;
		private void UpdateThumbnail()
		{
			if (updatelock == true || imagecache == null) return;
			updatelock = true;

			int w, h;
			GetThumbSize(out w, out h);

			// バックグラウンドで進行しているサムネイル作成を整理する
			int ch = ClientSize.Height - padding, cw = (ClientSize.Width - vsb.Width - padding / 2) / w;
			if (cw == 0) cw = 1;
			int v = vsb.Value;
			int y = -(v % h), s = v / h;
			if (s < 0) s = 0;
			int sno = s * cw;
			List<int> dellist = new List<int>();
			foreach (KeyValuePair<int, Thread> p in threaddata) {
				if (p.Key < sno || p.Key > sno + (cw * (ch / h))) {
					dellist.Add(p.Key);
				}
			}
			foreach (int i in dellist) {
				if (threaddata.ContainsKey(i) == true) threaddata[i].Abort();
			}

			// 横幅に何個のサムネイルが配置可能か
			vsb.Maximum = (imagecache.Length / cw + 1) * h - ClientSize.Height + textheight;
			if (vsb.Minimum < 0) {
				vsb.Minimum = 0;
				vsb.Value = 0;
				vsb.Enabled = false;
			} else {
				vsb.Enabled = true;
			}

			Invalidate();

			updatelock = false;
		}

		protected override void OnClientSizeChanged(EventArgs e)
		{
			if (this.IsHandleCreated == false) return;
			base.OnClientSizeChanged(e);
			UpdateThumbnail();
		}

		private void vsb_ValueChanged(object sender, EventArgs e)
		{
			if (this.IsHandleCreated == false) return;
			UpdateThumbnail();
		}

		private void tsSize_ValueChanged(object sender, EventArgs e)
		{
			UpdateThumbnail();
		}

		private Dictionary<int, Thread> threaddata = new Dictionary<int,Thread>();

		private void CreateThumbnail(int no)
		{
			if (threaddata.ContainsKey(no) == true) return;

			Thread thread = new Thread(new ParameterizedThreadStart(CreateThumbnailThread));
			thread.Start(no);
			threaddata.Add(no, thread);
		}

		private void CreateThumbnailThread(object o)
		{
			bool abort = false;
			int sno = (int)o;
			int index = imagecache[sno].index;
			IList<DBAttributedData> res;
			try {
			}catch(ThreadAbortException){
				abort = true;
			} finally {
				res = table.GetAttributedValue
					(DBAttrType.Thumbnail | DBAttrType.Title, index, index);
			}
			if (abort == true) return;
			try {
				// データベースからデータ取得
				if (res.Count > 0) {
					Image img = res[0].thumbnail;
					if (img == null && embed == true) {
						string path = (string)table.IFirebird.ExecScalar(
							string.Format("SELECT {0} FROM {1} WHERE {2}={3};",
							pct_name, table.TableName, Thumbnail.STR_ID, index));
						if(Path.IsPathRooted(path) == false){
							path = util.ToAbsolutePath(table.EditPath, path);
						}
						img = util.ExtractImage(path, new Size(256, 256));
					}
					imagecache[sno].loadfailed = (img == null);
					imagecache[sno].image = img;
					imagecache[sno].title = res[0].title;
				} else {
					imagecache[sno].loadfailed = true;
				}
				Invoke(new MethodInvoker(InvokeInvalidate));
			} catch (Exception) {
			} finally {
				threaddata.Remove(sno);
			}
		}

		private void InvokeInvalidate()
		{
			Invalidate();
		}

		private void DrawThumbnail(Graphics g)
		{
			if (imagecache == null || imagecache.Length == 0) return;

			g.SmoothingMode = SmoothingMode.AntiAlias;

			int w, h; GetThumbSize(out w, out h);
			Rectangle trc = new Rectangle(0, 0, tsTrackSize.Value, tsTrackSize.Value);
			Size isize = new Size(256, 256);

			StringFormat sf = StringFormat.GenericDefault;
			sf.LineAlignment = StringAlignment.Center;
			sf.Alignment = StringAlignment.Center;

			int ch = ClientSize.Height - padding, cw = (ClientSize.Width - vsb.Width - padding / 2) / w;
			if (cw == 0) cw = 1;
			int v = vsb.Value;
			int y = -(v % h), s = v / h;
			if (s < 0) s = 0;
			int sno = s * cw;
			if (sno >= imagecache.Length) return;
			while (y < ch) {
				Rectangle frc = trc;
				Rectangle rc = util.GetThumbnailRect(trc, isize);
				frc.Y = y + padding; rc.Y += y;
				for (int x = 0; x < cw; x++) {
					Image img = imagecache[sno].image;

					if (imagecache[sno].selected == true) {
						Rectangle arc = rc;
						arc.X += padding; arc.Width += margin * 2;
						arc.Y += padding; arc.Height += margin * 2;

						GraphicsPath path = new GraphicsPath();
						path.AddArc(new Rectangle(arc.Left, arc.Top, 6, 6), 180.0f, 90.0f);
						path.AddLine(arc.Left + 3, arc.Top, arc.Right - 3, arc.Top);
						path.AddArc(new Rectangle(arc.Right - 6, arc.Top, 6, 6), -90.0f, 90.0f);
						path.AddLine(arc.Right, arc.Top + 3, arc.Right, arc.Bottom - 3);
						path.AddArc(new Rectangle(arc.Right - 6, arc.Bottom - 6, 6, 6), 0.0f, 90.0f);
						path.AddLine(arc.Right - 3, arc.Bottom, arc.Left + 3, arc.Bottom);
						path.AddArc(new Rectangle(arc.Left, arc.Bottom - 6, 6, 6), 90.0f, 90.0f);
						//path.AddLine(arc.Left, arc.Bottom - 3, arc.Left, arc.Top - 3);
						path.CloseFigure();

						LinearGradientBrush gbrush = new LinearGradientBrush(arc,
							ControlPaint.LightLight(ControlPaint.LightLight(SystemColors.Highlight)),
							SystemColors.Highlight,
							LinearGradientMode.Vertical);
						g.FillPath(gbrush, path);
						g.DrawPath(SystemPens.Highlight, path);
						gbrush.Dispose();
					}

					if (img != null) {
						Rectangle prc = util.GetThumbnailRect(frc, img.Size);
						prc.X += margin + padding; prc.Y += margin;
						g.DrawImage(img, prc, new Rectangle(0, 0, img.Width, img.Height), GraphicsUnit.Pixel);
					} else if (imagecache[sno].loadfailed == true) {
						g.DrawString("No Image", this.Font, SystemBrushes.WindowText, rc, sf);
					} else {
						CreateThumbnail(sno);
					}

					string stitle = imagecache[sno].title;
					if (string.IsNullOrEmpty(stitle) == true) {
						stitle = string.Format("No.{0}", imagecache[sno].index);
					}
					g.DrawString(stitle, Font, SystemBrushes.WindowText,
						new Rectangle(frc.X + padding, frc.Bottom + padding, frc.Width, textheight), sf);

					/*
					g.DrawRectangle(Pens.Black, frc);
					g.DrawRectangle(Pens.Red, rc);
					g.DrawString(sno.ToString(), this.Font, Brushes.Black, frc);
					g.DrawRectangle(Pens.Blue, frc.X, frc.Bottom, frc.Width, textheight);
					*/

					sno++;
					frc.X += w; rc.X += w;
					if (sno >= imagecache.Length) goto NEXT;
				}
				y += h;
			}
		NEXT: ;
		}

		protected override void OnPaint(PaintEventArgs e)
		{
			base.OnPaint(e);

			Graphics g = e.Graphics;
			g.Clear(SystemColors.Window);

			if (tool.Enabled == false) {
				StringFormat sf = StringFormat.GenericDefault;
				sf.LineAlignment = StringAlignment.Center;
				sf.Alignment = StringAlignment.Center;
				Rectangle crc = ClientRectangle;
				crc.Width -= vsb.Width;
				g.DrawString(Properties.Resources.strNoPctAttr,
					this.Font, SystemBrushes.WindowText, crc, sf);
			} else {
				DrawThumbnail(g);
			}
		}

		protected override void OnSizeChanged(EventArgs e)
		{
			base.OnSizeChanged(e);
			Invalidate();
		}

		private void tsBtnQuick_Click(object sender, EventArgs e)
		{
			// クイック検索
			Dictionary<DBAttrType, string> d =
				table.GetAttributedParts(DBAttrType.Main | DBAttrType.Category | DBAttrType.Title);

			StringBuilder sb = new StringBuilder();
			sb.AppendFormat("SELECT {0} FROM {1} WHERE ", Thumbnail.STR_ID, table.TableName);

			bool first = true;
			FbFieldData[] fds = table.GetTableFields();
			foreach (KeyValuePair<DBAttrType, string> p in d) {
				foreach (FbFieldData fd in fds) {
					if (string.Compare(fd.fieldname, p.Value, true) == 0) {
						if (first == false) sb.Append(" OR "); else first = false;

						if (fd.type == "TEXT" || (fd.type == "BLOB" && fd.blobtype == 1)) {
							string text = tsTxtKeyword.Text;
							text = text.Replace("\\", "\\\\");
							text = text.Replace("%", "\\%");
							text = text.Replace("_", "\\_");
							sb.AppendFormat("{0} LIKE '%{1}%' ESCAPE '\\'", p.Value, text);
						} else {
							sb.AppendFormat("{0}={1}", p.Value, tsTxtKeyword.Text);
						}
					}
				}
			}
			sb.Append(";");

			BackgroundWorker w = new BackgroundWorker();
			w.DoWork += new DoWorkEventHandler(OnWorkGetThumbnailIndexes);
			w.RunWorkerCompleted += new RunWorkerCompletedEventHandler(OnCompleteGetThumbnailIndexes);
			w.RunWorkerAsync(sb.ToString());
		}

		private void OnWorkGetThumbnailIndexes(object sender, DoWorkEventArgs e)
		{
			List<ThumbnailData> newimgcache = new List<ThumbnailData>();
			datacache = new Dictionary<int,ThumbnailData>();
			IFbResult res = null;
			try {
				res = table.IFirebird.ExecCommand(e.Argument.ToString());
				foreach(Hashtable t in res){
					ThumbnailData data = new ThumbnailData();
					data.image = null;
					data.title = null;
					data.index = (int)t[Thumbnail.STR_ID];
					data.loadfailed = false;
					data.selected = false;
					newimgcache.Add(data);
					datacache.Add(data.index, data);
				}
			} finally {
				res.Close();
			}
			e.Result = newimgcache;
		}

		private void OnCompleteGetThumbnailIndexes(object sender, RunWorkerCompletedEventArgs e)
		{
			imagecache = ((List<ThumbnailData>)e.Result).ToArray();
			UpdateThumbnail();
		}

		private void tsShowAll_Click(object sender, EventArgs e)
		{
			BackgroundWorker w = new BackgroundWorker();
			w.DoWork += new DoWorkEventHandler(OnWorkGetThumbnailIndexes);
			w.RunWorkerCompleted += new RunWorkerCompletedEventHandler(OnCompleteGetThumbnailIndexes);
			w.RunWorkerAsync(string.Format("SELECT {0} FROM {1};", Thumbnail.STR_ID, table.TableName));
		}

		protected override bool ProcessDialogKey(Keys keyData)
		{
			switch(util.GetShortcutTypeName((Shortcut)(keyData | Control.ModifierKeys))){
			case "SelectAll":
				selectedindex.Clear();
				int n = 0;
				foreach (ThumbnailData d in imagecache) {
					d.selected = true;
					selectedindex.Add(n++);
				}
				Invalidate();
				break;
			default: return base.ProcessDialogKey(keyData);
			}
			return true;
		}
	}

	public class ThumbnailData
	{
		public int index;
		public Image image;
		public string title;
		public bool loadfailed;
		public bool selected;
	}
}
